@@ -5,6 +5,7 @@ The HAPServerHandler manages the state of the connection and handles incoming re
 The HAPSocket is a socket implementation that manages the "TLS" of the connection.
 """
 from http.server import HTTPServer, BaseHTTPRequestHandler
+from http import HTTPStatus
 import logging
 import socket
 import struct
@@ -94,6 +95,7 @@ class HAPServerHandler(BaseHTTPRequestHandler):
             "/pair-setup": "handle_pairing",
             "/pair-verify": "handle_pair_verify",
             "/pairings": "handle_pairings",
+            "/resource": "handle_resource",
         },
 
         "GET": {
@@ -504,21 +506,26 @@ class HAPServerHandler(BaseHTTPRequestHandler):
     def handle_set_characteristics(self):
         """Handles a client request to update certain characteristics."""
         if not self.is_encrypted:
-            raise UnprivilegedRequestException
+            logger.warning('Attemp to access unauthorised content from %s',
+                           self.client_address)
+            self.send_response(HTTPStatus.UNAUTHORIZED)
+            self.end_response(b'', close_connection=True)
 
-        # TODO: assert self.headers["authorization"] == state.pincode
-        data_len = int(self.headers["Content-Length"])
-        assert data_len > 0
+        data_len = int(self.headers['Content-Length'])
         requested_chars = json.loads(
-            self.rfile.read(data_len).decode("utf-8"))
+            self.rfile.read(data_len).decode('utf-8'))
+        logger.debug('Set characteristics content: %s', requested_chars)
 
-        chars = self.accessory_handler.set_characteristics(requested_chars,
-                                                           self.client_address)
-
-        data = json.dumps(chars).encode("utf-8")
-        self.send_response(207)
-        self.send_header("Content-Type", self.JSON_RESPONSE_TYPE)
-        self.end_response(data)
+        # TODO: Outline how chars return errors on set_chars.
+        try:
+            self.accessory_handler.set_characteristics(requested_chars,
+                                                       self.client_address)
+        except Exception as e:
+            logger.exception('Exception in set_characteristics: %s', e)
+            self.send_response(HTTPStatus.BAD_REQUEST)
+        else:
+            self.send_response(HTTPStatus.NO_CONTENT)
+            self.end_response(b'')
 
     def handle_pairings(self):
         """Handles a client request to update or remove a pairing."""
@@ -564,10 +571,26 @@ class HAPServerHandler(BaseHTTPRequestHandler):
         self.send_header("Content-Type", self.PAIRING_RESPONSE_TYPE)
         self.end_response(data)
 
+    def handle_resource(self):
+        """Get a snapshot from the camera."""
+        if not hasattr(self.accessory_handler.accessory, 'get_snapshot'):
+            raise ValueError('Got a request for snapshot, but the Accessory '
+                             'does not define a "get_snapshot" method')
+        data_len = int(self.headers['Content-Length'])
+        image_size = json.loads(
+                        self.rfile.read(data_len).decode('utf-8'))
+        image = self.accessory_handler.accessory.get_snapshot(image_size)
+        self.send_response(200)
+        self.send_header('Content-Type', 'image/jpeg')
+        self.end_response(image)
 
-class HAPSocket(socket.socket):
+
+class HAPSocket:
     """A socket implementing the HAP crypto. Just feed it as if it is a normal socket.
 
+    This implementation is something like a Proxy pattern - some calls to socket
+    methods are wrapped and some are forwarded as is.
+
     @note: HAP requires something like HTTP push. This implies we can have regular HTTP
     response and an outbound HTTP push at the same time on the same socket - a race
     condition. Thus, HAPSocket implements exclusive access to send and sendall to deal
@@ -582,20 +605,8 @@ class HAPSocket(socket.socket):
     IN_CIPHER_INFO = b"Control-Write-Encryption-Key"
 
     def __init__(self, sock, shared_key):
-        """Initialises this socket from the given socket."""
-        socket.socket.__init__(self, sock.family, sock.type, sock.proto, sock.fileno())
-        sock.detach()
-        # See if we are connected
-        try:
-            self.getpeername()
-        except OSError as e:
-            if e.errno != errno.ENOTCONN:
-                raise
-            self._connected = False
-        else:
-            self._connected = True
-
-        self._closed = False
+        """Initialise from the given socket."""
+        self.socket = sock
 
         self.shared_key = shared_key
         self.out_count = 0
@@ -613,6 +624,35 @@ class HAPSocket(socket.socket):
         self.num_in_recv = None  # Number of bytes received from the incoming block
         self.curr_in_block = None  # Bytes of the current incoming block
 
+    def __getattr__(self, attribute_name):
+        """Defer unknown behaviour to the socket"""
+        return getattr(self.socket, attribute_name)
+
+    def _get_io_refs(self):
+        """Get `socket._io_refs`."""
+        return self.socket._io_refs
+
+    def _set_io_refs(self, value):
+        """Set `socket._io_refs`."""
+        self.socket._io_refs = value
+
+    _io_refs = property(_get_io_refs, _set_io_refs)
+    """`socket.makefile` uses a `SocketIO` to wrap the socket stream. Internally,
+    this uses `socket._io_refs` directly to determine if a socket object needs to be
+    closed when its FileIO object is closed.
+
+    Because `_io_refs` is assigned as part of this process, it bypasses getattr. To get
+    around this, let's make _io_refs our property and proxy calls to the socket.
+    """
+
+    def makefile(self, *args, **kwargs):
+        """Return a file object that reads/writes to this object.
+
+        We need to implement this, otherwise the socket's makefile will use the socket
+        object and we won't en/decrypt.
+        """
+        return socket.socket.makefile(self, *args, **kwargs)
+
     def _set_ciphers(self):
         """Generate out/inbound encryption keys and initialise respective ciphers."""
         outgoing_key = hap_hkdf(self.shared_key, self.CIPHER_SALT, self.OUT_CIPHER_INFO)
@@ -655,7 +695,7 @@ class HAPSocket(socket.socket):
                     # It may be that we already read some data and we have
                     # 1 byte left, return whatever we have.
                     return result
-                block_length_bytes = socket.socket.recv(self, self.LENGTH_LENGTH)
+                block_length_bytes = self.socket.recv(self.LENGTH_LENGTH)
                 if not block_length_bytes:
                     return result
                 # TODO: handle this
@@ -669,9 +709,8 @@ class HAPSocket(socket.socket):
                 buflen -= self.LENGTH_LENGTH
             else:
                 # Read as much from the current block as possible.
-                part = socket.socket.recv(self,
-                                          min(buflen,
-                                              self.curr_in_total - self.num_in_recv))
+                part = self.socket.recv(min(buflen,
+                                            self.curr_in_total - self.num_in_recv))
                 # Check what is actually received
                 actual_len = len(part)
                 self.curr_in_block += part
@@ -717,7 +756,7 @@ class HAPSocket(socket.socket):
             offset += length
             self.out_count += 1
             result += ciphertext
-        socket.socket.sendall(self, result)
+        self.socket.sendall(result)
         return total
 
 
@@ -741,7 +780,7 @@ class HAPServer(socketserver.ThreadingMixIn,
                      b"Content-Length: "
 
     TIMEOUT_ERRNO_CODES = (errno.ECONNRESET, errno.EPIPE, errno.EHOSTUNREACH,
-                           errno.ETIMEDOUT)
+                           errno.ETIMEDOUT, errno.EHOSTDOWN)
 
     @classmethod
     def create_hap_event(cls, bytesdata):
@@ -804,11 +843,11 @@ class HAPServer(socketserver.ThreadingMixIn,
 
     def server_close(self):
         """Close all connections."""
-        logger.info("Stopping HAP server")
-        super(HAPServer, self).server_close()
+        logger.info('Stopping HAP server')
         for sock in self.connections.values():
             self._close_socket(sock)
         self.connections.clear()
+        super().server_close()
 
     def push_event(self, bytesdata, client_addr):
         """Send an event to the current connection with the provided data.
